const require = (await import('node:module')).createRequire(import.meta.url); const __chunk_filename = (await import('node:url')).fileURLToPath(import.meta.url); const __chunk_dirname = (await import('node:path')).dirname(__chunk_filename);
import {
  require_dist
} from "./chunk-RJTRUG2J.js";
import "./chunk-34MYV7JD.js";
export default require_dist();
